from execute import execute_listing_query
from collections import defaultdict
from memoize import memoize

office_name_cache = {}

FIXED_NWMLS_LOOKUPS = {
    'ST': {
        'A': 'Active',
        'CT': 'Contingent',
        'PB': 'Pending BU Requested',
        'PF': 'Pending Feasability',
        'PI': 'Pending Inspection',
        'PS': 'Pending Short Sale',
        'P': 'Pending',
        'E': 'Expired',
        'T': 'Temp. Off Markt.',
        'SFR': 'Sale Fail Release',
        'CA': 'Cancelled',
        'R': 'Rented',
        'S': 'Sold',
    },
    'SD': {
        '00':'Not Known',
        'ABR':'Aberdeen',
        'AD':'Adna',
        'ALMI':'Almira',
        'ANA':'Anacortes',
        'AR':'Arlington',
        'ASOT':'Asotin-Anatone',
        'AUB':'Auburn',
        'BE':'Bellevue',
        'BEL':'Bellingham',
        'BENG':'Benge',
        'BG':'Battle Ground',
        'BICK':'Bickleton',
        'BIS':'Bainbridge Island',
        'BLN':'Blaine',
        'BOI':'Boistfort',
        'BRG':'Bridgeport',
        'BRI':'Brinnon',
        'BRL':'Burlington',
        'BRM':'Bremerton #100c',
        'BRW':'Brewster',
        'BTH':'Bethel',
        'CAM':'Camas',
        'CAS':'Cashmere',
        'CEN':'Centralia',
        'CENT':'Centerville',
        'CENV':'Central Valley',
        'CHE':'Chehalis',
        'CHEN':'Cheney',
        'CHEW':'Chewelah',
        'CHI':'Chimacum #49',
        'CK':'Central Kitsap #401',
        'CLAR':'Clarkston',
        'CLE':'Cle Elum-Roslyn',
        'CLH':'Coulee-Hartline',
        'CLP':'Clover Park',
        'COFX':'Colfax',
        'COLB':'Columbia - Burbank',
        'COLH':'Columbia - Hunters',
        'COLP':'College Place',
        'COLT':'Colton',
        'COLV':'Colville',
        'CON':'Concrete',
        'COS':'Cosmopolis',
        'CPF':'Cape Flattery',
        'CPV':'Coupeville',
        'CRB':'Carbonado',
        'CRE':'Creston',
        'CRES':'Crescent',
        'CRK':'Castle Rock',
        'CSC':'Cascade',
        'CURL':'Curlew',
        'CUSI':'Cusick',
        'CWY':'Conway',
        'DAM':'Damman',
        'DAR':'Darrington',
        'DAVE':'Davenport',
        'DAYT':'Dayton',
        'DGR':'Dieringer',
        'DIXI':'Dixie',
        'DPRK':'Deer Park',
        'EAS':'Easton',
        'EAT':'Eatonville',
        'ED':'Edmonds',
        'ELM':'Elma',
        'ELN':'Ellensburg',
        'EMT':'Eastmont',
        'ENDI':'Endicott',
        'ENM':'Enumclaw',
        'ENT':'Entiat',
        'EPH':'Ephrata',
        'EV':'Everett',
        'EVA':'Evaline',
        'EVAL':'East Valley Spokane',
        'EVG':'Evergreen',
        'EVGR':'Evergreen - Hunters',
        'EYAK':'East Valley Yakima',
        'FED':'Federal Way',
        'FIF':'Fife',
        'FINL':'Finley',
        'FPS':'Franklin Pierce',
        'FREE':'Freeman',
        'FRN':'Ferndale',
        'GARF':'Garfield',
        'GF':'Granite Falls',
        'GLEN':'Glenwood',
        'GMT':'Green Mountain',
        'GOLD':'Goldendale',
        'GRC':'Grand Coulee',
        'GRF':'Griffin',
        'GRGR':'Granger',
        'GRNO':'Great Northern',
        'GRP':'Grapeview #54',
        'GRVW':'Grandview',
        'HARR':'Harrington',
        'HC':'Hood Canal #404',
        'HGL':'Highline',
        'HLND':'Highland',
        'HOC':'Hockinson',
        'HOQ':'Hoquiam',
        'INCH':'Inchelium',
        'IND':'Index',
        'ISS':'Issaquah',
        'KAHL':'Kahlotus',
        'KAL':'Kalama',
        'KEL':'Kelso',
        'KELR':'Keller',
        'KENN':'Kennewick',
        'KETF':'Kettle Falls',
        'KION':'Kiona-Benton City',
        'KLIC':'Klickitat',
        'KNT':'Kent',
        'KTS':'Kittitas',
        'LAC':'La Conner',
        'LACS':'Lacrosse',
        'LAMO':'Lamont',
        'LC':'La Center',
        'LCH':'Lake Chelan',
        'LGV':'Longview',
        'LIBR':'Liberty',
        'LIND':'Lind',
        'LKD':'Lakewood',
        'LKW':'Lake Washington',
        'LOON':'Loon Lake',
        'LPZ':'Lopez Island',
        'LS':'Lake Stevens',
        'LYLE':'Lyle',
        'LYN':'Lynden',
        'MABT':'Mabton',
        'MAR':'Marysville',
        'MARW':'Mary Walker',
        'MCC':'McCleary',
        'MEAD':'Mead',
        'MEDL':'Medical Lake',
        'MER':'Meridian',
        'MET':'Methow Valley',
        'MIS':'Mercer Island',
        'MK':'Mary M. Knight #311',
        'MLA':'Mill A',
        'MLK':'Moses Lake',
        'MNS':'Manson',
        'MON':'Monroe',
        'MOR':'Morton',
        'MOS':'Mossyrock',
        'MOT':'Montesano',
        'MSF':'Mansfield',
        'MTAD':'Mount Adams',
        'MTB':'Mount Baker',
        'MTP':'Mount Pleasant',
        'MTV':'Mount Vernon',
        'MUK':'Mukilteo',
        'NAP':'Napavine',
        'NAS':'Naselle-Grays River',
        'NAVY':'Naches Valley',
        'NBC':'North Beach',
        'NES':'Nespelem',
        'NEWP':'Newport',
        'NFRA':'North Franklin',
        'NK':'North Kitsap #400',
        'NM':'North Mason #403',
        'NMFS':'Nine Mile Falls',
        'NOO':'Nooksack Valley',
        'NPRT':'Northport',
        'NRV':'North River',
        'NTH':'Northshore',
        'NTN':'North Thurston',
        'OAK':'Oakville',
        'OAKS':'Oakesdale',
        'OCB':'Ocean Beach',
        'OCO':'Ocosta',
        'ODES':'Odessa',
        'OH':'Oak Harbor',
        'OKA':'Okanogan',
        'OLY':'Olympia',
        'OMA':'Omak',
        'ONA':'Onalaska',
        'ONCR':'Onion Creek',
        'ORC':'Orcas Island',
        'ORCH':'Orchard Prairie',
        'ORIE':'Orient',
        'ORO':'Orondo',
        'ORT':'Orting',
        'OTHE':'Othello',
        'OTHR':'Other',
        'OVL':'Oroville',
        'PAL':'Palisades',
        'PALO':'Palouse',
        'PASC':'Pasco',
        'PAT':'Pateros',
        'PATS':'Paterson',
        'PE':'Pe Ell',
        'PEN':'Peninsula',
        'PI':'Pioneer #402',
        'POME':'Pomeroy',
        'PRES':'Prescott',
        'PROS':'Prosser',
        'PTA':'Port Angeles',
        'PTT':'Port Townsend #50',
        'PULL':'Pullman',
        'PUY':'Puyallup',
        'QNC':'Quincy',
        'QTS':'Queets-Clearwater',
        'QUI':'Quinault',
        'QUL':'Quilcene',
        'QUT':'Quillayute',
        'RAI':'Rainier',
        'RAY':'Raymond',
        'REAR':'Reardan-Edwall',
        'REN':'Renton',
        'REPU':'Republic',
        'RICH':'Richland',
        'RID':'Ridgefield',
        'RITZ':'Ritzville',
        'ROC':'Rochester',
        'ROOS':'Roosevelt',
        'ROSA':'Rosalia',
        'RSID':'Riverside',
        'RVW':'Riverview',
        'RYL':'Royal',
        'SAN':'San Juan Island',
        'SAT':'Satsop',
        'SB':'South Bend',
        'SDW':'Sedro Woolley',
        'SEA':'Seattle',
        'SELA':'Selah',
        'SELK':'Selkirk',
        'SEQ':'Sequim',
        'SHL':'Shelton',
        'SHW':'Shaw Island',
        'SK':'South Kitsap',
        'SKM':'Skamania',
        'SKY':'Skykomish',
        'SNO':'Snohomish',
        'SPL':'Soap Lake',
        'SPOK':'Spokane',
        'SPRA':'Sprague',
        'SRI':'Shoreline',
        'SS':'South Side #42',
        'SSID':'Sunnyside',
        'ST':'Stanwood',
        'STAR':'Star',
        'STBU':'Starbuck',
        'STE':'Steilacoom Historica',
        'STEP':'Steptoe',
        'STH':'Stehekin',
        'STJO':'St. John',
        'STV':'Stevenson-Carson',
        'SUL':'Sultan',
        'SUM':'Sumner',
        'SUMM':'Summit Valley',
        'SVS':'Snoqualmie Valley',
        'SWI':'South Whidbey Island',
        'TAC':'Tacoma',
        'TEKO':'Tekoa',
        'TEN':'Tenino',
        'THL':'Taholah',
        'THO':'Thorp',
        'THS':'Tahoma',
        'TOL':'Toledo',
        'TOPP':'Toppenish',
        'TOU':'Toutle Lake',
        'TOUC':'Touchet',
        'TROU':'Trout Lake',
        'TSK':'Tonasket',
        'TUK':'Tukwila',
        'TUM':'Tumwater',
        'UGAP':'Union Gap',
        'UNP':'University Place',
        'VALY':'Valley',
        'VAN':'Vancouver',
        'VDR':'Vader',
        'VSH':'Vashon Island',
        'WAH':'Wahkiakum',
        'WAIT':'Waitsburg',
        'WAPA':'Wapato',
        'WAS':'Washougal',
        'WAWA':'Walla Walla',
        'WDL':'Woodland',
        'WELL':'Wellpinit',
        'WEN':'Wenatchee',
        'WHR':'White River',
        'WHT':'White Pass',
        'WILB':'Wilbur',
        'WIN':'Winlock',
        'WIS':'Wishkah',
        'WISH':'Wishram',
        'WLC':'Wilson Creek',
        'WLP':'Willapa',
        'WLU':'Wahluke',
        'WRD':'Warden',
        'WSAL':'White Salmon',
        'WTUC':'Washtucna',
        'WV':'Waterville',
        'WVAL':'West Valley Spokane',
        'WYAK':'West Valley Yakima',
        'YAKI':'Yakima',
        'YEL':'Yelm',
        'ZILA':'Zillah',
    }
}

@memoize
def _get_offices(username, password):
    for row in execute_listing_query(username, password, 'RetrieveOfficeData', {'MLS': 'nwmls'}):
        yield dict([(c.tag.replace('{http://www.nwmls.com/Schemas/Standard/StandardXML1_1.xsd}', ''), c.text) for c in row.getchildren()])


@memoize
def _get_amenities(username, password, prop_type):
    for r in execute_listing_query(username, password, 'RetrieveAmenityData', {'MLS': 'nwmls', 'PropertyType': prop_type}):
        field_name = r.find('{http://www.nwmls.com/Schemas/General/EverNetAmenityXML.xsd}Code').text
        values = r.find('{http://www.nwmls.com/Schemas/General/EverNetAmenityXML.xsd}Values').getchildren()
        row = dict([(c.tag.replace('{http://www.nwmls.com/Schemas/General/EverNetAmenityXML.xsd}', ''), c.text) for c in values])
        yield field_name, row


def _get_office_table(offices):
    return dict([(office.get('{http://www.nwmls.com/Schemas/General/EverNetOfficeXML.xsd}OfficeMLSID'), office.get('{http://www.nwmls.com/Schemas/General/EverNetOfficeXML.xsd}OfficeName')) for office in offices])


def _get_amenities_table(amenities):
    field_map = defaultdict(dict)
    for field_name, row in amenities:
        field_map[field_name][row['Code']] = row['Description']
    return field_map


def _look_up_dynamic_fields(username, password, row):
    prop_type = row.get('PTYP')
    out = {}
    if not prop_type:
        return out
    lookup_table = _get_amenities_table(_get_amenities(username, password, prop_type))
    for key, value in row.iteritems():
        if not value:
            continue
        if key not in lookup_table:
            out[key] = value
        else:
            out[key] = []
            for v in value.split('|'):
                out[key].append(lookup_table[key].get(v))
    return out

def _look_up_fixed_fields(row):
    out = {}
    for key, value in row.iteritems():
        if key in FIXED_NWMLS_LOOKUPS:
            out[key] = FIXED_NWMLS_LOOKUPS[key].get(value, value)
        else:
            out[key] = value
    return out

def _add_office_name_field(row, username, password):
    offices = _get_office_table(_get_offices(username, password))
    row['OFFICENAME'] = offices.get(row.get('LO'))
    return row


def look_up_all_fields(username, password, row):
    row = _look_up_dynamic_fields(username, password, row)
    row = _add_office_name_field(row, username, password)
    row = _look_up_fixed_fields(row)
    return row



from itertools import tee
from types import GeneratorType

Tee = tee([], 1)[0].__class__

def memoize(f):
    cache={}
    def ret(*args):
        if args not in cache:
            cache[args]=f(*args)
        if isinstance(cache[args], (GeneratorType, Tee)):
            # the original can't be used any more,
            # so we need to change the cache as well
            cache[args], r = tee(cache[args])
            return r
        return cache[args]
    return ret



import itertools

ALL_PROPERTY_TYPES = [
    "BUSO",
    "COMI",
    "COND",
    "FARM",
    "MANU",
    "MULT",
    "RENT",
    "RESI",
    "TSHR",
    "VACL",
]


def iterate_criteria(begin_date, end_date, property_types=None, areas=None, cities=None, status=None):
    """
    NOTE: If you don't pass criteria to Evernet, it will assume you only want residential properties.
    That's why I will iterate through all availabe types unless you pass me a list of types you want
    explicitly.
    """
    base_criteria = {
        'MLS': 'nwmls',
        'BeginDate': begin_date,
        'EndDate': end_date,
    }
    areas = areas or [None]
    cities = cities or [None]
    property_types = property_types or ALL_PROPERTY_TYPES
    status = status or [None]
    for property_type, area, city, status in itertools.product(property_types, areas, cities, status):
        criteria = {}
        if area:
            criteria['Area'] = area
        if city:
            criteria['City'] = city
        if status:
            criteria['Status'] = status
        if property_type:
            criteria['PropertyType'] = property_type
        criteria.update(base_criteria)
        yield criteria



VALID_LISTING_QUERY_FIELDS = [
    'MLS',
    'ListingNumber',
    'PropertyType',
    'Status',
    'County',
    'Area',
    'City',
    'BeginDate',
    'EndDate',
    'OfficeId',
    'AgentId',
    'Bedrooms',
    'Bathrooms',
]


PHOTO_QUERY = '<ImageQuery xmlns="NWMLS:EverNet:ImageQuery:1.0"><Auth><UserId>{UserId}</UserId><Password>{Password}</Password></Auth><Query><ByListingNumber>{ListingNumber}</ByListingNumber></Query><Results><Schema>NWMLS:EverNet:ImageData:1.0</Schema></Results></ImageQuery>'


LISTING_QUERY = '<EverNetQuerySpecification xmlns="urn:www.nwmls.com/Schemas/General/EverNetQueryXML.xsd"><Message><Head><UserId>{UserId}</UserId><Password>{Password}</Password><SchemaName>StandardXML1_2</SchemaName></Head><Body><Query>{Query}</Query><Filter>{Filter}</Filter></Body></Message></EverNetQuerySpecification>'

def _validate_query(query, valid_fields):
    for i, v in query.iteritems():
        if i not in valid_fields:
            raise ValueError("Invalid field %s" % i)


def _build_query(query):
    tag = '<{tag}>{value}</{tag}>'
    return ''.join([tag.format(tag=i, value=v) for i, v in query.items()])


def listing_query(username, password, query, filter=''):
    _validate_query(query, VALID_LISTING_QUERY_FIELDS)

    return LISTING_QUERY.format(
        UserId=username,
        Password=password,
        Query=_build_query(query),
        Filter=filter
    )


def photo_query(username, password, listing_number):

    return PHOTO_QUERY.format(
        UserId=username,
        Password=password,
        ListingNumber=listing_number,
    )



import base64
import datetime
from execute import execute_listing_query, execute_photo_query
from lookups import look_up_all_fields
from criteria import iterate_criteria



def get_all_listings(username, password, property_types=[], areas=[], cities=[], status=[]):
    return get_new_listings(
        username,
        password,
        hours_previous=876581,
        property_types=property_types,
        areas=areas,
        cities=cities,
        status=status,
    )


def get_mls_numbers(username, password, property_types=[], areas=[], cities=[], status=['A'], date_min=None, date_max=None):
    if not date_min:
        date_min = datetime.datetime(year=1990, month=1, day=1, hour=0, minute=0, second=0)
    if not date_max:
        date_max = datetime.datetime(year=3000, month=1, day=1, hour=0, minute=0, second=0)
    date_min = date_min.strftime('%Y-%m-%dT%H:%M:%S')
    date_max = date_max.strftime('%Y-%m-%dT%H:%M:%S')
    for criteria in iterate_criteria(date_min, date_max, status=status, property_types=property_types, areas=areas, cities=cities):
        for row in execute_listing_query(username, password, 'RetrieveListingData', criteria, filter="LN,ST"):
            info = dict([(c.tag.replace('{http://www.nwmls.com/Schemas/Standard/StandardXML1_2.xsd}', ''), c.text) for c in row.getchildren()])
            yield info.get('LN')


def get_new_listings(username, password, hours_previous=24, property_types=[], areas=[], cities=[], status=[]):
    begin_date = (datetime.datetime.utcnow() - datetime.timedelta(hours=hours_previous)).strftime('%Y-%m-%dT%H:%M:%S')
    end_date = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S')

    for criteria in iterate_criteria(begin_date, end_date, property_types, areas, cities, status):
        for row in execute_listing_query(username, password, 'RetrieveListingData', criteria):
            out = dict([(c.tag.replace('{http://www.nwmls.com/Schemas/Standard/StandardXML1_2.xsd}', ''), c.text) for c in row.getchildren()])
            yield look_up_all_fields(username, password, out)


def get_photos(username, password, listing_id):
    for row in execute_photo_query(username, password, listing_id):
        out = dict([(c.tag.replace('{NWMLS:EverNet:ImageData:1.0}', ''), c.text) for c in row.getchildren()])
        out['BLOB'] = base64.b64decode(out['BLOB'])
        yield out

def get_property(username, password, mls_number):
    criteria = {
        'MLS':'nwmls',
        'ListingNumber':mls_number
    }
    results = execute_listing_query(username, password, 'RetrieveListingData', criteria)
    first_result = next(results) # only return first match
    out = dict([(c.tag.replace('{http://www.nwmls.com/Schemas/Standard/StandardXML1_2.xsd}', ''), c.text) for c in first_result.getchildren()])
    return look_up_all_fields(username, password, out)

